#![feature(random)]
#![feature(inline_const_pat)]

use std::io::{ErrorKind, prelude::*};
use std::net::{TcpListener, TcpStream};
use std::random;
use std::sync::{Arc, Mutex};
use std::thread::{self, JoinHandle, sleep};
use clap::Parser;

#[derive(Parser, Debug)]
#[command(version, about, long_about = None)]
struct Args {
    /// Port to bind on.
    #[arg(short, long, default_value_t = 7867)]
    port: u16,
}

#[derive(Clone, PartialEq, Eq)]
enum UserStatus {
    ONLINE,
    DISCONNECTED,
    IDLE,
    CUSTOM(String),
}

impl UserStatus {
    fn get_status_string(&self) -> String {
        match self {
            UserStatus::ONLINE => "\x1b[0m\x1b[32monline\x1b[0m".to_string(),
            UserStatus::IDLE => "\x1b[0m\x1b[33midle\x1b[0m".to_string(),
            UserStatus::DISCONNECTED => "\x1b[0m\x1b[31mdisconnected\x1b[0m".to_string(),
            UserStatus::CUSTOM(val) => format!("\x1b[0m\x1b[34m{}\x1b[0m", val),
        }
    }
}

#[derive(Clone)]
struct User {
    name: String,
    status: UserStatus,
    token: String,
}

#[derive(Clone)]
struct Message {
    sender: User,
    contents: String,
    timestamp: time::OffsetDateTime,
    me_message: bool,
}

type Channel = Arc<Mutex<UnwrappedChannel>>;

#[derive(Clone)]
struct UnwrappedChannel {
    name: String,
    messages: Vec<Message>,
    online_users: Vec<User>,
}

fn handle_client(mut stream: TcpStream, channel: Channel) {
    stream
        .write_all(b"\x1b[33mEnter your chosen nickname(up to 30 characters):\x1b[0m\x1b[4m\n")
        .unwrap();
    let mut buf = [0u8; 30];
    let mut i = 0usize;
    let mut buf1 = [0u8];
    'bufreadloop: while buf1[0] != b'\n' || i == 0 {
        stream.read_exact(&mut buf1).unwrap();
        if buf1[0] == b'\n' && i == 0 {
            stream.write_all(
                b"\x1b[0m\x1b[31;1mEnter at least one character.\n\x1b[0m\x1b[33mEnter your chosen nickname(up to 30 characters):\x1b[0m\x1b[4m\n"
            ).unwrap();
            buf = [0u8; 30];
            buf1[0] = b'\0';
            continue;
        } else if buf1[0] == b'\n' && i >= 30 {
            stream.write_all(
                b"\x1b[0m\x1b[31;1mEnter less than 30 characters.\n\x1b[0m\x1b[33mEnter your chosen nickname(up to 30 characters):\x1b[0m\x1b[4m\n"
            ).unwrap();
            buf = [0u8; 30];
            buf1[0] = b'\0';
            i = 0;
            continue;
        } else if buf1[0] == b'\n' {
            let new_name = String::from_utf8_lossy(&(buf[..i])).to_string();
            if new_name.to_lowercase() == "netbot" {
                stream.write_all(
                    b"\x1b[0m\x1b[31;1mNetbot: Hey! Rude.\n \
                    \x1b[0m\x1b[33mEnter your chosen nickname(up to 30 characters):\x1b[0m\x1b[4m\n"
                ).unwrap();
                buf = [0u8; 30];
                buf1[0] = b'\0';
                i = 0;
                continue 'bufreadloop;
            }
            for user in &channel.lock().unwrap().online_users {
                if user.name == new_name {
                    if user.status == UserStatus::DISCONNECTED {
                        stream
                            .write_all(b"\x1b[0m\x1b[33mEnter your login token(up to 50 characters):\x1b[0m\x1b[4m\n")
                            .unwrap();
                        let mut buf = [0u8; 50];
                        let mut i = 0usize;
                        let mut buf1 = [0u8];
                        'bufreadloop2: while buf1[0] != b'\n' || i == 0 {
                            stream.read_exact(&mut buf1).unwrap();
                            if buf1[0] == b'\n' && i >= 50 {
                                stream.write_all(
                                    b"\x1b[0m\x1b[31;1mEnter less than 50 characters.\n \
                                    \x1b[0m\x1b[33mEnter your login token(up to 50 characters):\x1b[0m\x1b[4m\n"
                                ).unwrap();
                                buf = [0u8; 50];
                                buf1[0] = b'\0';
                                i = 0;
                                continue;
                            } else if buf1[0] == b'\n' {
                                if user.token != String::from_utf8_lossy(&(buf[..i])) {
                                    break 'bufreadloop2;
                                }
                                break 'bufreadloop;
                            }
                            if i < 50 {
                                buf[i] = buf1[0];
                                i += 1;
                            }
                        }
                    }
                    stream.write_all(
                        b"\x1b[0m\x1b[31;1mSorry, but that nickname is taken.\n \
                        \x1b[0m\x1b[33mEnter your chosen nickname(up to 30 characters):\x1b[0m\x1b[4m\n"
                    ).unwrap();
                    buf = [0u8; 30];
                    buf1[0] = b'\0';
                    i = 0;
                    continue 'bufreadloop;
                }
            }
            break;
        }
        if i < 30 {
            buf[i] = buf1[0];
            i += 1;
        }
    }

    stream
        .write_fmt(format_args!(
            "\x1b[0m\x1b[32;1mWelcome, {}! Joining server \"{}\".\x1b[0m\n",
            String::from_utf8_lossy(&(buf[..i])),
            channel.lock().unwrap().name
        ))
        .unwrap();

    let mut user = User {
        name: String::from_utf8_lossy(&(buf[..i])).to_string(),
        status: UserStatus::ONLINE,
        token: format!("{:x}", random::random::<u128>()),
    };

    for usr in &mut channel.lock().unwrap().online_users {
        if usr.name == user.name {
            usr.status = user.status.clone();
            usr.token = user.token.clone();
            break;
        }
    }

    stream
        .write_fmt(format_args!(
            "\x1b[0m\x1b[32;1mYour token to rejoin after quitting is {}.\x1b[0m\n",
            user.token
        ))
        .unwrap();

    let netbot = User {
        name: "Netbot".to_owned(),
        status: UserStatus::ONLINE,
        token: "".to_owned(),
    };

    channel.lock().unwrap().online_users.push(user.clone());

    let (recieve_terminate_send, recieve_terminate_recv) = std::sync::mpsc::channel::<()>();
    let recieve_thread: JoinHandle<_>;

    {
        let channel = channel.clone();
        let mut stream = stream.try_clone().unwrap();

        let time_format = time::format_description::well_known::Rfc2822;

        recieve_thread = thread::spawn(move || {
            let mut current_message_count = channel.lock().unwrap().messages.len();
            loop {
                if let Ok(_) = recieve_terminate_recv.try_recv() {
                    stream.shutdown(std::net::Shutdown::Both).unwrap();
                    panic!("recieve thread exiting");
                }
                let new_message_count = channel.lock().unwrap().messages.len();
                if current_message_count != new_message_count {
                    let new_messages = &channel.lock().unwrap().messages[current_message_count..];
                    for msg in new_messages {
                        if !msg.me_message {
                            stream
                                .write_fmt(format_args!(
                                    "\x1b[0m\x1b[34m[{}]\x1b[0m \x1b[1m{}\x1b[0m \x1b[31m⇛\x1b[0m \x1b[35m{}\x1b[0m\n",
                                    msg.timestamp.format(&time_format).unwrap(),
                                    msg.sender.name,
                                    msg.contents
                                ))
                                .unwrap();
                        } else {
                            stream
                                .write_fmt(format_args!(
                                    "\x1b[0m\x1b[34m[{}]\x1b[0m \x1b[1;3m{}\x1b[0m \x1b[3m{}\x1b[0m\n",
                                    msg.timestamp.format(&time_format).unwrap(),
                                    msg.sender.name,
                                    msg.contents
                                ))
                                .unwrap();
                        }
                    }

                    current_message_count = new_message_count;
                }

                // so that not all the threads will sync at once
                let randomval = random::random::<u64>() / (18446744073709551615u64 / 20);
                sleep(std::time::Duration::from_millis(240 + randomval));
            }
        });
    }

    channel.lock().unwrap().messages.push(Message {
        sender: netbot.clone(),
        contents: format!("User {} has joined!", user.name),
        timestamp: time::OffsetDateTime::now_local().unwrap(),
        me_message: false,
    });

    let mut buf = Vec::<u8>::new();
    let mut buf1 = [0u8];

    loop {
        if let Err(err) = stream.read_exact(&mut buf1) {
            if err.kind() == ErrorKind::UnexpectedEof {
                user.status = UserStatus::DISCONNECTED;
                for usr in &mut channel.lock().unwrap().online_users {
                    if usr.name == user.name {
                        usr.status = user.status.clone();
                        break;
                    }
                }

                channel.lock().unwrap().messages.push(Message {
                    sender: netbot.clone(),
                    contents: format!("User {} has left.", user.name),
                    timestamp: time::OffsetDateTime::now_local().unwrap(),
                    me_message: false,
                });

                recieve_terminate_send.send(()).unwrap();
                let _ = recieve_thread.join();
                return;
            }
        }

        if buf1[0] == b'\x08' {
            buf.pop();
        }

        if buf1[0] == b'\n' && buf.len() > 0 {
            let contents = String::from_utf8_lossy(&buf).to_string();
            if !contents.starts_with("/") {
                channel.lock().unwrap().messages.push(Message {
                    sender: user.clone(),
                    contents,
                    timestamp: time::OffsetDateTime::now_local().unwrap(),
                    me_message: false,
                });
            } else {
                if contents.split(" ").collect::<Vec<&str>>()[0] == "/help"
                    || contents.split(" ").collect::<Vec<&str>>()[0] == "/?"
                {
                    stream
                        .write_all(
                            b"\x1b[0m\x1b[32mValid commands:\n \
                                /help, /?: Print this help.\n \
                                /status [online|idle|disconnected|(any value)]: Set your status.\n \
                                /user [nick]: Get information about a user.\n \
                                /server [new name]: Get or set the server name.\n \
                                /users: List users.\n \
                                /me x: Say that you are doing x(for example, \"/me purrs\" as netbot would result in the output of \"netbot purrs\").\x1b[0m\n",
                        )
                        .unwrap();
                } else if contents.split(" ").collect::<Vec<&str>>()[0] == "/me" {
                    if contents.split(" ").collect::<Vec<&str>>().len() < 2
                        || contents.split(" ").collect::<Vec<&str>>()[1] == ""
                    {
                        stream
                            .write_fmt(format_args!(
                                "\x1b[0m\x1b[31mYou need to provide at least one argument to /me!\n\x1b[0m",
                            ))
                            .unwrap();
                    } else {
                        channel.lock().unwrap().messages.push(Message {
                            sender: user.clone(),
                            contents: contents.replace("/me ", ""),
                            timestamp: time::OffsetDateTime::now_local().unwrap(),
                            me_message: true,
                        });
                    }
                } else if contents.split(" ").collect::<Vec<&str>>()[0] == "/status"
                    || contents == "/user"
                {
                    let contents = contents.split(" ").collect::<Vec<&str>>();
                    if contents.len() == 1 {
                        stream
                            .write_fmt(format_args!(
                                "\x1b[0m\x1b[32mYou are currently marked as {}\n\x1b[0m",
                                user.clone().status.get_status_string()
                            ))
                            .unwrap();
                    } else {
                        match contents[1] {
                            "online" => {
                                user.status = UserStatus::ONLINE;
                            }
                            "idle" => {
                                user.status = UserStatus::IDLE;
                            }
                            "disconnected" => {
                                user.status = UserStatus::DISCONNECTED;
                            }
                            other => {
                                user.status = UserStatus::CUSTOM(other.to_owned());
                            }
                        }
                        for usr in &mut channel.lock().unwrap().online_users {
                            if usr.name == user.name {
                                usr.status = user.status.clone();
                            }
                        }
                        stream
                            .write_fmt(format_args!(
                                "\x1b[0m\x1b[32mYou are now marked as {}\n\x1b[0m",
                                user.status.get_status_string()
                            ))
                            .unwrap();
                    }
                } else if contents.split(" ").collect::<Vec<&str>>()[0] == "/server" {
                    let contents = contents.split(" ").collect::<Vec<&str>>();
                    if contents.len() == 1 {
                        stream
                            .write_fmt(format_args!(
                                "\x1b[0m\x1b[32mCurrent server name: {}.\n\x1b[0m",
                                channel.lock().unwrap().name
                            ))
                            .unwrap();
                    } else {
                        channel.lock().unwrap().name = contents[1].to_owned();
                        stream
                            .write_fmt(format_args!(
                                "\x1b[0m\x1b[32mNew server name: {}.\n\x1b[0m",
                                channel.lock().unwrap().name
                            ))
                            .unwrap();
                    }
                } else if contents.split(" ").collect::<Vec<&str>>()[0] == "/user" {
                    let contents = contents.split(" ").collect::<Vec<&str>>();
                    let mut found_user = false;
                    for usr in &mut channel.lock().unwrap().online_users {
                        if usr.name == contents[1] {
                            stream
                                .write_fmt(format_args!(
                                    "\x1b[0m\x1b[32mUser {} is currently {}.\n\x1b[0m",
                                    usr.name,
                                    usr.status.get_status_string()
                                ))
                                .unwrap();
                            found_user = true;
                            break;
                        }
                    }
                    if !found_user {
                        stream
                            .write_fmt(format_args!(
                                "\x1b[0m\x1b[32mUnknown user \"{}\".\n\x1b[0m",
                                contents[1]
                            ))
                            .unwrap();
                    }
                } else if contents.starts_with("/users") {
                    stream.write_all(b"\x1b[0m\x1b[32mUsers:\x1b[0m").unwrap();
                    for usr in &channel.lock().unwrap().online_users {
                        stream
                            .write_fmt(format_args!(
                                "\n{} (currently {})",
                                usr.name,
                                usr.status.get_status_string()
                            ))
                            .unwrap();
                    }
                    stream.write_all(b"\n\x1b[0m").unwrap();
                } else {
                    stream
                        .write_all(b"\x1b[0m\x1b[31;1mThat's an invalid command!\n\x1b[0m")
                        .unwrap();
                }
            }
            buf.clear();
        }
        if buf1[0] == b'\n' {
            continue;
        }
        buf.push(buf1[0]);
    }
}

fn main() -> std::io::Result<()> {
    let args = Args::parse();

    let listener = TcpListener::bind(format!("0.0.0.0:{}", args.port))?;
    let channel: Channel = Arc::new(Mutex::new(UnwrappedChannel {
        name: "Rawnetchat".to_string(),
        messages: vec![],
        online_users: vec![],
    }));

    for stream in listener.incoming() {
        let channel: Channel = channel.clone();
        thread::spawn(|| {
            handle_client(stream.unwrap(), channel);
        });
    }

    Ok(())
}
